(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Copyright (C) 2004, 2005 Milenko "DCoder" Mitrovic                       *
 *  Mail: dcoder@dsp-worx.de                                                 *
 *  Web:  http://www.dsp-worx.de                                             *
 *                                                                           *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

unit ImageFilter;

interface

uses
  BaseClass, DirectShow9, DSUtil, Windows, ActiveX, SysUtils,
  Graphics, Registry;

const
  CLSID_ImageSource: TGUID = '{C8B48845-CA82-4C7B-9230-76B56E554EDF}';

  sudPinTypes: TRegPinTypes = (
    clsMajorType: @MEDIATYPE_Video;
    clsMinorType: @MEDIASUBTYPE_NULL
  );

  sudOutputPin: TRegFilterPins = (
    strName: 'Output';
    bRendered: FALSE;
    bOutput: TRUE;
    bZero: FALSE;
    bMany: FALSE;
    oFilter: nil;
    strConnectsToPin: nil;
    nMediaTypes: 1;
    lpMediaType: @sudPinTypes
  );

  UNITS = 10000000;

  FPS_30 = UNITS div 30;
  FPS_20 = UNITS div 20;
  FPS_10 = UNITS div 10;
  FPS_5 = UNITS div 5;
  FPS_4 = UNITS div 4;
  FPS_3 = UNITS div 3;
  FPS_2 = UNITS div 2;
  FPS_1 = UNITS div 1;

type
  TBCImageSourcePin = class(TBCSourceStream)
  protected
    FSharedState: TBCCritSec;
    FFrameLength: Integer;
    FFrameNumber: Int64;
    FBitmap: TBitmap;
  public
    constructor Create(out hr: HResult; Filter: TBCSource);
    destructor Destroy; override;

    function LoadImage(AFileName: WideString): Boolean;

    function GetMediaType(MediaType: PAMMediaType): HResult; override;
    function CheckMediaType(MediaType: PAMMediaType): HRESULT; override;
    function DecideBufferSize(Allocator: IMemAllocator; Properties: PAllocatorProperties): HRESULT; override;
    function FillBuffer(Sample: IMediaSample): HResult; override;
    function Notify(Filter: IBaseFilter; q: TQuality): HRESULT; override;  stdcall;
  end;

  TBCImageSource = class(TBCSource, IFileSourceFilter, IPersist)
  private
    FPin: TBCImageSourcePin;
    FFileName: WideString;
  public
    constructor Create; overload;
    constructor Create(ObjName: string; Unk: IUnKnown; out hr: HRESULT); overload;
    constructor CreateFromFactory(Factory: TBCClassFactory; const Controller: IUnknown); override;
    destructor Destroy; override;
    // IFileSourceFilter
    function Load(pszFileName: PWCHAR; const pmt: PAMMediaType): HResult; stdcall;
    function GetCurFile(out ppszFileName: PWideChar; pmt: PAMMediaType): HResult; stdcall;
  end;

  function DllRegisterServer: HResult; stdcall;
  function DllUnregisterServer: HResult; stdcall;

implementation

uses
  GraphicEx;

Type
  TColor24 =
  Record
    cBlue  : Byte;
    cGreen : Byte;
    cRed   : Byte;
  End;
  TColor32 =
  Record
    cRed   : Byte;
    cGreen : Byte;
    cBlue  : Byte;
    cAlpha : Byte;
  End;
  PColor24          = ^TColor24;
  PColor32          = ^TColor32;
  TMyScanLine       = Array[0..4095] of TColor24;
  TMyScanLine32     = Array[0..4095] of TColor;
  TMyScanLineGrid   = Array[0..4095] of ^TMyScanLine;
  TMyScanLine32Grid = Array[0..4095] of ^TMyScanLine32;

{$DEFINE COL32}

// Bilinear Resize Code by Blight (www.inmatrix.com)

procedure ImageBilinearResize(SrcBitmap,DestBitmap : TBitmap; XInOffset,YInOffset,SrcWidth,SrcHeight,XOutOffset,YOutOffset,DestWidth,DestHeight : Integer);
var
  x,y,xP,yP      : Integer;
  yP2,xP2        :  Integer;
  Read,Read2     : ^TMyScanLine;
  Read32,Read232 : ^TMyScanLine32;
  t,z,z2,iz2     : Integer;
  pc             : PColor24;
  pc32           : PColor32;
  w1,w2,w3,w4    : Integer;
  Col1,Col2      : PColor24;
  Col132,Col232  : PColor32;
  P              : ^TMyScanLine;
  PL             : ^TMyScanLineGrid;
  P32            : ^TMyScanLine32;
  PL32           : ^TMyScanLine32Grid;
  PDif           : Integer;
  ypShr          : Integer;
  BitAccu        : Integer;

begin
  If (DestWidth<=2) or (DestHeight<=2) then Exit;

  xP2  := ((SrcWidth -1) shl 11) div DestWidth;
  yP2  := ((SrcHeight-1) shl 11) div DestHeight;
  yP   := 0;

  {$IFNDEF COL32}
  New(PL);

  // Calculate memory used for each line display for quick scanline seeks (dest image)
  PL^[0] := SrcBitmap.Scanline[YInOffset];
  PL^[1] := SrcBitmap.Scanline[YInOffset+1];
  PDif   := Integer(PL^[1])-Integer(PL^[0]);

  // Pre-Calculate scanline positions for source image
  For Y := 2 to SrcHeight-1 do Integer(PL^[Y]) := Integer(PL^[Y-1])+PDif;

  P    := DestBitmap.Scanline[YOutOffset];
  PDif := Integer(DestBitmap.Scanline[YOutOffset+1])-Integer(P);
  For Y := 0 to DestHeight-1 do
  Begin
    xP   := 0;
    ypShr:= yP shr 11;
    Pointer(Read) := Pointer(PL^[ypShr]);
    If ypShr < SrcHeight-1 then Pointer(Read2) := Pointer(PL^[ypShr+1]) else
      Pointer(Read2) := Pointer(PL^[ypShr]);

    pc  := @P[XOutOffset];
    z2  := yP and 2047;
    iz2 := 2048 - z2;
    For X := 0 to DestWidth-1 do
    Begin
      t    := xP shr 11;
      Col1 := @Read[XInOffset+t];
      Col2 := @Read2[XInOffset+t];
      z    := xP and 2047;
      w2   := (z * iz2) shr 11;
      w1   := iz2 - w2;
      w4   := (z * z2) shr 11;
      w3   := z2-w4;
      pc^.cBlue  := (Col1.cBlue  * w1 + PColor24(Integer(Col1) + 3).cBlue  * w2+
                     Col2.cBlue  * w3 + PColor24(Integer(Col2) + 3).cBlue  * w4) shr 11;
      pc^.cGreen := (Col1.cGreen * w1 + PColor24(Integer(Col1) + 3).cGreen * w2+
                     Col2.cGreen * w3 + PColor24(Integer(Col2) + 3).cGreen * w4) shr 11;
      pc^.cRed   := (Col1.cRed   * w1 + PColor24(Integer(Col1) + 3).cRed   * w2+
                     Col2.cRed   * w3 + PColor24(Integer(Col2) + 3).cRed   * w4) shr 11;
      Inc(pc);
      Inc(xP,xP2);
    End;
    Inc(yP,yP2);
    Inc(Integer(P),PDif);
  End;
  Dispose(PL);
  {$ELSE}
  New(PL32);
  // Calculate memory used for each line display for quick scanline seeks (dest image)
  PL32^[0] := SrcBitmap.Scanline[YInOffset];
  PL32^[1] := SrcBitmap.Scanline[YInOffset+1];
  PDif     := Integer(PL32^[1])-Integer(PL32^[0]);

  // Pre-Calculate scanline positions for source image
  For Y := 2 to SrcHeight-1 do Integer(PL32^[Y]) := Integer(PL32^[Y-1])+PDif;

  P32  := DestBitmap.Scanline[YOutOffset];
  PDif := Integer(DestBitmap.Scanline[YOutOffset+1])-Integer(P32);
  For Y := 0 to DestHeight-1 do
  Begin
    xP   := 0;
    ypShr:= yP shr 11;
    Pointer(Read32) := Pointer(PL32^[ypShr]);
    If ypShr < SrcHeight-1 then Pointer(Read232) := Pointer(PL32^[ypShr+1]) else
      Pointer(Read232) := Pointer(PL32^[ypShr]);

    pc32 := @P32[XOutOffset];
    z2  := yP and 2047;
    iz2 := 2048 - z2;
    For X := 0 to DestWidth-1 do
    Begin
      t      := xP shr 11;
      Col132 := @Read32[XInOffset+t];
      Col232 := @Read232[XInOffset+t];
      z      := xP and 2047;
      w2     := (z * iz2) shr 11;
      w1     := iz2 - w2;
      w4     := (z * z2) shr 11;
      w3     := z2-w4;
      pc32^.cBlue  := (Col132.cBlue  * w1 + PColor32(Integer(Col132) + 4).cBlue  * w2+
                       Col232.cBlue  * w3 + PColor32(Integer(Col232) + 4).cBlue  * w4) shr 11;
      pc32^.cGreen := (Col132.cGreen * w1 + PColor32(Integer(Col132) + 4).cGreen * w2+
                       Col232.cGreen * w3 + PColor32(Integer(Col232) + 4).cGreen * w4) shr 11;
      pc32^.cRed   := (Col132.cRed   * w1 + PColor32(Integer(Col132) + 4).cRed   * w2+
                       Col232.cRed   * w3 + PColor32(Integer(Col232) + 4).cRed   * w4) shr 11;
      Inc(pc32);
      Inc(xP,xP2);
    End;
    Inc(yP,yP2);
    Inc(Integer(P32),PDif);
  End;
  Dispose(PL32);
  {$ENDIF}
end;

(*** TBCImageSourcePin ********************************************************)

constructor TBCImageSourcePin.Create(out hr: HResult; Filter: TBCSource);
begin
  inherited Create('', hr, Filter, 'Out');
  FSharedState := TBCCritSec.Create;
  FFrameLength := FPS_1;
  FFrameNumber := 0;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf32bit;
end;

destructor TBCImageSourcePin.Destroy;
begin
  FBitmap.Free;
  FreeAndNil(FSharedState);
  inherited Destroy;;
end;

function TBCImageSourcePin.GetMediaType(MediaType: PAMMediaType): HResult;
begin
  FFilter.StateLock.Lock;
  try
    if (MediaType = nil) then
    begin
      Result := E_POINTER;
      Exit;
    end;

    MediaType.cbFormat := SizeOf(TVideoInfoHeader);
    MediaType.pbFormat := CoTaskMemAlloc(MediaType.cbFormat);
    if (MediaType.pbFormat = nil) then
    begin
      Result := E_OUTOFMEMORY;
      Exit;
    end;

    ZeroMemory(MediaType.pbFormat, MediaType.cbFormat);

    with PVideoInfoHeader(MediaType.pbFormat)^ do
    begin
      bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
      bmiHeader.biWidth := FBitmap.Width;
      bmiHeader.biHeight := FBitmap.Height;

      bmiHeader.biPlanes := 1;
      bmiHeader.biBitCount := 32;
      bmiHeader.biCompression := BI_RGB;
      bmiHeader.biSizeImage := GetBitmapSize(@bmiHeader);
      bmiHeader.biXPelsPerMeter := 0;
      bmiHeader.biYPelsPerMeter := 0;
      bmiHeader.biClrUsed := 0;
      bmiHeader.biClrImportant := 0;

      MediaType.lSampleSize := bmiHeader.biSizeImage;
      dwBitRate := 0;
      dwBitErrorRate := 0;
      SetRectEmpty(rcSource);
      SetRectEmpty(rcTarget);
      AvgTimePerFrame := FFrameLength;
    end;

    MediaType.majortype := MEDIATYPE_Video;
    MediaType.subtype := MEDIASUBTYPE_RGB32;
    MediaType.formattype := FORMAT_VideoInfo;
    MediaType.bTemporalCompression := False;
    MediaType.bFixedSizeSamples := True;

    Result := S_OK;

  finally
    FFilter.StateLock.UnLock;
  end;
end;

function TBCImageSourcePin.CheckMediaType(MediaType: PAMMediaType): HRESULT;
begin
  Result := E_INVALIDARG;

  if not IsEqualGUID(MediaType.majortype, MEDIATYPE_Video)
    then Exit;
  if not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_RGB32)
    then Exit;
  if not IsEqualGUID(MediaType.formattype, FORMAT_VideoInfo)
    then Exit;
  if not MediaType.bFixedSizeSamples
    then Exit;
  if MediaType.pbFormat = nil
    then Exit;

  with PVideoInfoHeader(MediaType.pbFormat)^ do
  begin
    if (bmiHeader.biWidth < FBitmap.Width) or
       (abs(bmiHeader.biHeight) < FBitmap.Height) or
       (bmiHeader.biBitCount <> 32)
       then Exit;
  end;

  Result := S_OK;
end;

function TBCImageSourcePin.DecideBufferSize(Allocator: IMemAllocator; Properties: PAllocatorProperties): HRESULT;
var
  Actual: ALLOCATOR_PROPERTIES;
begin
  if (Allocator = nil) or (Properties = nil) then
  begin
    Result := E_POINTER;
    Exit;
  end;

  FFilter.StateLock.Lock;
  try
    if (Properties.cBuffers = 0)
      then Properties.cBuffers := 2;

    Properties.cbBuffer := PVideoInfoHeader(AMMediaType.pbFormat).bmiHeader.biSizeImage;

    Result := Allocator.SetProperties(Properties^, Actual);
    if Failed(Result)
      then Exit;

    if (Actual.cbBuffer < Properties.cbBuffer)
      then Result := E_FAIL
      else Result := S_OK;

  finally
    FFilter.StateLock.UnLock;
  end;
end;

function TBCImageSourcePin.FillBuffer(Sample: IMediaSample): HResult;
var
  buffer: PByte;
  size: Integer;
  Start, Stop: REFERENCE_TIME;
  pmt: PAMMediaType;
  width: Integer;
  new_width: Integer;
  i: Integer;
begin
  if (Sample = nil) then
  begin
    Result := E_POINTER;
    Exit;
  end;

  if sample.GetMediaType(pmt) = S_OK then
  begin
    SetMediaType(pmt);
    DeleteMediaType(pmt);
  end;

  with PVideoInfoHeader(Fmt.pbFormat)^ do
  begin
    width := FBitmap.Width;
    new_width := bmiHeader.biWidth;
  end;

  FSharedState.Lock;
  try
    size := (FBitmap.Height * width) shl 2;
    Sample.GetPointer(PByte(buffer));
    Sample.SetActualDataLength(size);

    if PVideoInfoHeader(Fmt.pbFormat).bmiHeader.biHeight < 0 then
    begin
      for i := 0 to FBitmap.Height -1 do
      begin
        Move(PByte(FBitmap.ScanLine[i])^, buffer^, width shl 2);
        inc(buffer, new_width shl 2);
      end;
    end else
    begin
      for i := FBitmap.Height -1 downto 0 do
      begin
        Move(PByte(FBitmap.ScanLine[i])^, buffer^, width shl 2);
        inc(buffer, new_width shl 2);
      end;
    end;

    Start := FFrameNumber * FFrameLength;
    Stop := Start + FFrameLength;

    Sample.SetTime(@Start, @Stop);
    Inc(FFrameNumber);

    Sample.SetSyncPoint(True);

    Result := S_OK;
  finally
    FSharedState.UnLock;
  end;
end;

function TBCImageSourcePin.Notify(Filter: IBaseFilter; q: TQuality): HRESULT;
begin
  Result := S_OK;
end;

function TBCImageSourcePin.LoadImage(AFileName: WideString): Boolean;
const
  MAX_SIZE = 2000;
var
  picture: TPicture;
  bitmap: TBitmap;
  w, h: Integer;
  ratio: Double;
begin
  Result := False;

  picture := TPicture.Create;
  try
   picture.LoadFromFile(AFileName);
  except
    picture.Free;
    Exit;
  end;


  w := picture.Width;
  h := picture.Height;

  if (w > MAX_SIZE) or (h > MAX_SIZE) then
  begin
    if w > h
      then ratio := w / MAX_SIZE
      else ratio := h / MAX_SIZE;

    bitmap := TBitmap.Create;

    bitmap.PixelFormat := pf32bit;
    bitmap.Width := picture.Width;
    bitmap.Height := picture.Height;
    bitmap.Canvas.Draw(0, 0, picture.Graphic);

    FBitmap.Width := Round(picture.Width / ratio);
    FBitmap.Height := Round(picture.Height / ratio);

    ImageBilinearResize(bitmap, FBitmap, 0, 0, bitmap.Width, bitmap.Height, 0, 0, FBitmap.Width, FBitmap.Height);

    bitmap.Free;
  end else
  begin
    FBitmap.Width := picture.Width;
    FBitmap.Height := picture.Height;
    FBitmap.Canvas.Draw(0, 0, picture.Graphic);
  end;

  picture.Free;
  Result := (FBitmap.Width > 0) and (FBitmap.Height > 0) and (FBitmap.PixelFormat = pf32bit);
end;

(*** TBCImageSource ***********************************************************)

constructor TBCImageSource.Create;
var
  hr: HRESULT;
begin
  inherited Create('TBCImageSource', nil, CLSID_ImageSource);

  FPin := TBCImageSourcePin.Create(hr, Self);
  if (hr <> S_OK) and (FPin = nil)
    then hr := E_OUTOFMEMORY;
  FFileName := '';
end;

constructor TBCImageSource.Create(ObjName: string; Unk: IUnKnown; out hr: HRESULT);
begin
  inherited Create(ObjName, Unk, CLSID_ImageSource);

  FPin := TBCImageSourcePin.Create(hr, Self);
  if (hr <> S_OK) and (FPin = nil)
    then hr := E_OUTOFMEMORY;
  FFileName := '';
end;

constructor TBCImageSource.CreateFromFactory(Factory: TBCClassFactory; const Controller: IUnknown);
var
  hr: HRESULT;
begin
  Create(Factory.Name, Controller, hr);
end;

destructor TBCImageSource.Destroy;
begin
  FreeAndNil(FPin);
  inherited;
end;

function TBCImageSource.Load(pszFileName: PWCHAR; const pmt: PAMMediaType): HResult;
begin
  Result := E_FAIL;

  if not Assigned(pszFileName)
    then Exit;

  if (GetFileAttributesW(pszFileName) = $FFFFFFFF)
    then Exit;

  if not FPin.LoadImage(pszFileName)
    then Exit;

  FFileName := pszFileName;
  Result := S_OK;
end;

function TBCImageSource.GetCurFile(out ppszFileName: PWideChar; pmt: PAMMediaType): HResult;
begin
  if not Assigned(@ppszFileName) then
  begin
    Result := E_POINTER;
    Exit;
  end;

  if (FFileName  = '') then
  begin
    ppszFileName := nil;
    Result := E_FAIL;
    Exit;
  end;

  AMGetWideString(FFileName, ppszFileName);
  Result := S_OK;
end;

(*** Register/Unregister ******************************************************)

const
  EXTENSIONS: array[0..39] of String = (
    'bmp', 'ico', 'wmf', 'emf', 'jfif', 'jpg', 'jpe', 'jpeg', 'rle', 'dib',
    'win', 'vst', 'vda', 'tga', 'icb', 'tiff', 'tif', 'fax', 'eps', 'pcx', 'pcc',
    'scr', 'rpf', 'rla', 'sgi', 'rgba', 'rgb', 'bw', 'psd', 'pdd', 'ppm',
    'pgm', 'pbm', 'cel', 'pic', 'pcd', 'gif', 'cut', 'psp', 'png'
  );

function DllRegisterServer: HResult; stdcall;
var
  i: Integer;
begin
  // Setup DirectShow to use our Filesource for ImageExtensions Extensions
  with TRegistry.Create do
  begin
    RootKey := HKEY_CLASSES_ROOT;
    for i := 0 to Length(EXTENSIONS) -1 do
    begin
      if OpenKey('Media Type\Extensions\.' + EXTENSIONS[i] ,True) then
      begin
        WriteString('Source Filter', GUIDToString(CLSID_ImageSource));
        CloseKey;
      end;
    end;
    Free;
  end;
  Result := BaseClass.DllRegisterServer;
end;

const
  CLSID_MS_ImageSource: TGuid = '{7DF62B50-6843-11D2-9EEB-006008039E37}';
  OLD_EXTENSIONS: array[0..4] of String = ( 'bmp', 'jpe', 'jpg', 'jpeg', 'tga' );

function DllUnregisterServer: HResult; stdcall;
var
  i: Integer;
begin
  with TRegistry.Create do
  begin
    RootKey := HKEY_CLASSES_ROOT;
    for i := 0 to Length(EXTENSIONS) -1 do
    begin
      if KeyExists('Media Type\Extensions\.' + EXTENSIONS[i])
        then DeleteKey('Media Type\Extensions\.' + EXTENSIONS[i]);
    end;
    for i := 0 to Length(OLD_EXTENSIONS) -1 do
    begin
      if OpenKey('Media Type\Extensions\.' + OLD_EXTENSIONS[i] ,True) then
      begin
        WriteString('Source Filter', GUIDToString(CLSID_MS_ImageSource));
        CloseKey;
      end;
    end;
    Free;
  end;
  Result := BaseClass.DllUnregisterServer;
end;

initialization

  TBCClassFactory.CreateFilter(
    TBCImageSource, 'DC-Image Source', CLSID_ImageSource, CLSID_LegacyAmFilterCategory,
    MERIT_DO_NOT_USE, 1, @sudOutputPin
  );

end.

